#
#  feedback_func.tcl -- Feedback control support functions.
#
#  Written by William Lavender.
#
#------------------------------------------------------------------------
#
#  Copyright 2004-2005 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc create_message_log_window { label } {

	set window .message_log

	frame $window
	pack $window -expand yes -fill both

	frame $window.top
	pack $window.top -side top -expand yes -fill x

	label $window.top.left_filler -text "           "
	label $window.top.label -text "Message Log"

	button $window.top.exit_button -text "Exit Program" \
			-bg lightblue -activebackground lightblue \
			-command exit

	grid $window.top.left_filler -row 0 -column 0 -padx 20 -sticky w
	grid $window.top.label       -row 0 -column 1 -padx 10
	grid $window.top.exit_button -row 0 -column 2 -padx 10 -sticky e

	grid columnconfigure $window.top 0 -weight 1
	grid columnconfigure $window.top 1 -weight 1
	grid columnconfigure $window.top 2 -weight 1

        #
        #  Create the log window.
        #

        iwidgets::scrolledtext $window.log

        $window.log configure -visibleitems 56x10
        $window.log configure -hscrollmode dynamic
        $window.log configure -vscrollmode dynamic

        pack $window.log -expand yes -fill both -side top

        #
        #  Install handlers to redirect the output of mx_info(), mx_error()
        #  and MX_DEBUG() to this window.
        #

	Mx::set_all_output_functions feedback_control_output_function

	#
	#  Install handler for mx_info_dialog().
	#

	Mx::set_info_dialog_function feedback_control_info_dialog_function

	log_puts "*** $label Initializing ***"
}

proc feedback_format { format_specifier value } {

	if [ catch { format $format_specifier $value } result ] {
		set result $value
	}
	return $result
}

proc feedback_control_output_function { message } {
	log_puts $message
	update
}

proc feedback_control_info_dialog_function { text_prompt 
						gui_prompt button_label } {
	bell
	tk_dialog .feedback_control_info_dialog "MX Info" \
		$gui_prompt info 0 $button_label
}

proc feedback_control_user_interrupt_function {} {

	update

	set interrupt_occurred $FeedbackControl::user_interrupt

	set FeedbackControl::user_interrupt 0

	return $interrupt_occurred
}

proc log_puts { message } {
	set textwindow [ .message_log.log childsite ]

	$textwindow configure -state normal

	.message_log.log insert end "$message\n"

	$textwindow configure -state disabled

	.message_log.log yview moveto 1.0
}

#############################################################################

proc setup_feedback_control_database { motor_database } {

	#
	# Create a new record list.
	#

	set record_list [ Mx::RecordList ::mxrecordlist_#auto ]

	#
	# Enable plotting.
	#

	$record_list set_plot_enable 1

	#
	# Read the motor and scan databases.
	#

	$record_list read_database $motor_database

	#
	# Finish initializing MX.
	#

	$record_list finish_database_initialization
	$record_list initialize_hardware 0

	return $record_list
}

proc create_feedback_control_gui { record_list bpm } {

	set main_window .main

	frame $main_window
	pack $main_window -before .message_log

	create_feedback_control_main_window $record_list $bpm $main_window

	setup_feedback_control_callbacks $record_list $bpm
}

proc find_callback_record { record_list item_name record_name } {

	global callback_records

	set status [ catch { $record_list get_record $record_name } \
			record ]

	if { $status != 0 } {
		log_puts \
		"Error: '$record_name' record not found in database."

		return -code error
	}

	set callback_records($item_name) $record
}

proc add_feedback_control_callback { callback } {
	global callback_list

	lappend callback_list $callback
}

proc delete_feedback_control_callback { callback } {
	global callback_list

	set i [ lsearch -exact $callback_list $callback ]

	if { $i < 0 } {
		set t \
	    "Attempt to delete callback '$callback' from callback list failed."

		Mx::error_message "Callback Deletion Failed." $t "" ""

		return
	}

	set callback_list [ lreplace $callback_list $i $i ]
}

proc start_feedback_control_callbacks { } {

	set FeedbackControl::callback_interval 1000	;# in milliseconds

	after 1 [ list process_feedback_control_callbacks ]
}

proc process_feedback_control_callbacks { } {

	global callback_list
	global callback_records
	global callback_values

#	log_puts "process_feedback_control_callbacks invoked"

	catch {

            foreach name [ array names callback_records ] {

                set record $callback_records($name)

                set superclass [ $record get_field "mx_superclass" ]

                if { $superclass == "variable" } {
            	    if [ catch { $record receive_variable } \
            					callback_values($name) ] \
            	    {
            		set callback_values($name) "*****"
            	    }
                } elseif { $superclass == "device" } {

            	    set class [ $record get_field "mx_class" ]

            	    if { $class == "motor" } {
            	        if [ catch { $record get_position } \
            					callback_values($name) ] \
            	        {
            		    set callback_values($name) "*****"
            	        }

            	        set busy_name "${name}_busy"

            	        if [ catch { $record is_busy } \
            					callback_values($busy_name) ] \
            	        {
            		    set callback_values($busy_name) "*****"
            	        }
		    } elseif { $class == "amplifier" } {

			set gain_name "${name}_gain"

			if [ catch { $record get_gain } \
					callback_values($gain_name) ] \
			{
			    set callback_values($gain_name) "*****"
			}

			set offset_name "${name}_offset"

			if [ catch { $record get_offset } \
					callback_values($offset_name) ] \
			{
			    set callback_values($offset_name) "*****"
			}
            	    } elseif { $class == "analog_input" } {
            	        if [ catch { $record read } callback_values($name) ] {
            		    set callback_values($name) "*****"
            	        }
            	    } elseif { $class == "analog_output" } {
            	        if [ catch { $record read } callback_values($name) ] {
            		    set callback_values($name) "*****"
            	        }
            	    } elseif { $class == "digital_input" } {
            	        if [ catch { $record read } callback_values($name) ] {
            		    set callback_values($name) "*****"
            	        }
            	    } elseif { $class == "digital_output" } {
            	        if [ catch { $record read } callback_values($name) ] {
            		    set callback_values($name) "*****"
            	        }
            	    } elseif { $class == "relay" } {
            	        if [ catch { $record get_relay_status } \
            					callback_values($name) ] {
            		    set callback_values($name) "*****"
            	        }
		    } elseif { $class == "scaler" } {
		        if [ catch { $record read } callback_values($name) ] {
			    set callback_values($name) "*****"
		        }
            	    } else {
            	        set callback_values($name) "+++++"
            	    }
                } else {
            	    set callback_values($name) "+++++"
                }

                #
                # process_feedback_control_callbacks can take a very long time ( ~1 sec )
                # to execute.  Thus, it is vitally important that the 'update'
                # command be invoked periodically so that the GUI seems more
                # responsive to the user.
                #
                # Apparently, 'update idletasks' is not sufficient for this
                # since feedback_control never becomes idle during the time that 
                # process_feedback_control_callbacks is running.
                #

                update
            }

            if [ catch { energy_from_lambda $callback_values(lambda) } \
            					callback_values(energy) ] {
            	set callback_values(energy) "!!!!!"
            }

            foreach callback $callback_list {
            	$callback
            }

	}

	after $FeedbackControl::callback_interval \
		[ list process_feedback_control_callbacks ]
}

#
#  Create the main window.
#

proc create_feedback_control_main_window { record_list bpm window } {

	set FeedbackControl::main_window $window

	#-----

	set frame $window.bpm_frame

	frame $frame
	pack $frame -side top

	create_bpm_frame $record_list $frame

	#-----

	frame $window.sep0 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep0 -expand yes -fill x -pady 5 -side top

	set frame $window.sp_frame

	frame $frame
	pack $frame -side top

	create_setpoint_frame $record_list $frame

	#-----

	frame $window.sep1 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep1 -expand yes -fill x -pady 5 -side top

	set frame $window.servo_frame

	frame $frame
	pack $frame -side top

	create_servo_frame $record_list $frame

	#-----

	frame $window.sep1a -height 2 -borderwidth 1 -relief sunken
	pack $window.sep1a -expand yes -fill x -pady 5 -side top

	set frame $window.param_frame

	frame $frame
	pack $frame -side top

	create_param_frame $record_list $frame

	#-----


	frame $window.sep1b -height 2 -borderwidth 1 -relief sunken
	pack $window.sep1b -expand yes -fill x -pady 5 -side top

	set frame $window.limits_frame

	frame $frame
	pack $frame -side top

	create_limits_frame $record_list $frame

	#-----

	frame $window.sep2 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep2 -expand yes -fill x -pady 5 -side top

	set frame $window.qbpm_control_frame

	frame $frame
	pack $frame -side top

	create_qbpm_control_frame $record_list $bpm $frame

	frame $window.sep3 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep3 -expand yes -fill x -pady 5 -side top

	add_feedback_control_callback main_window_callback
}

proc main_window_callback {} {

	global callback_values

#	log_puts "main_window_callback invoked"

	set window $FeedbackControl::main_window

	update_bpm_frame $window.bpm_frame $FeedbackControl::bpm

	update_setpoint_frame $window.sp_frame

	update_servo_frame $window.servo_frame

	update_param_frame $window.param_frame

	update_limits_frame $window.limits_frame

	update_qbpm_control_frame $window.qbpm_control_frame
}

proc feedback_control_format { format_specifier value } {

	if [ catch { format $format_specifier $value } result ] {
		set result $value
	}
	return $result
}

#############################################################################

proc setup_feedback_control_callbacks { record_list bpm } {

	find_callback_record $record_list "mjack_pitch" "mjack_pitch"
	find_callback_record $record_list "chi"   	"chi"

	find_callback_record $record_list "bpm_posx" "${bpm}_posx"
	find_callback_record $record_list "bpm_posy" "${bpm}_posy"

	find_callback_record $record_list "bpm" $bpm

	find_callback_record $record_list "bpm_gx" "${bpm}_gx"
	find_callback_record $record_list "bpm_gy" "${bpm}_gy"
	find_callback_record $record_list "bpm_a1" "${bpm}_a1"
	find_callback_record $record_list "bpm_a2" "${bpm}_a2"
	find_callback_record $record_list "bpm_b1" "${bpm}_b1"
	find_callback_record $record_list "bpm_b2" "${bpm}_b2"
	find_callback_record $record_list "bpm_c1" "${bpm}_c1"
	find_callback_record $record_list "bpm_c2" "${bpm}_c2"
	find_callback_record $record_list "bpm_d1" "${bpm}_d1"
	find_callback_record $record_list "bpm_d2" "${bpm}_d2"
	find_callback_record $record_list "bpm_average" "${bpm}_average"
	find_callback_record $record_list "num_qbpm_reads" "num_qbpm_reads"

	find_callback_record $record_list "bpm_sum" "${bpm}_sum"
	find_callback_record $record_list "bpm_curr1" "${bpm}_curr1"
	find_callback_record $record_list "bpm_curr2" "${bpm}_curr2"
	find_callback_record $record_list "bpm_curr3" "${bpm}_curr3"
	find_callback_record $record_list "bpm_curr4" "${bpm}_curr4"

	find_callback_record $record_list "bpm_x_setpoint" "bpm_x_setpoint"
	find_callback_record $record_list "bpm_y_setpoint" "bpm_y_setpoint"

	find_callback_record $record_list "bpm_x_params" "bpm_x_params"
	find_callback_record $record_list "bpm_y_params" "bpm_y_params"

	find_callback_record $record_list "bpm_sum_min" "bpm_sum_min"
	find_callback_record $record_list "bpm_x_min" "bpm_x_min"
	find_callback_record $record_list "bpm_x_max" "bpm_x_max"
	find_callback_record $record_list "bpm_x_deadband" "bpm_x_deadband"
	find_callback_record $record_list "bpm_y_min" "bpm_y_min"
	find_callback_record $record_list "bpm_y_max" "bpm_y_max"
	find_callback_record $record_list "bpm_y_deadband" "bpm_y_deadband"

	find_callback_record $record_list "x_fb_enable" "x_fb_enable"
	find_callback_record $record_list "y_fb_enable" "y_fb_enable"

	find_callback_record $record_list "bpm_x_kp" "bpm_x_kp"
	find_callback_record $record_list "bpm_x_ki" "bpm_x_ki"
	find_callback_record $record_list "bpm_x_kd" "bpm_x_kd"

	find_callback_record $record_list "bpm_y_kp" "bpm_y_kp"
	find_callback_record $record_list "bpm_y_ki" "bpm_y_ki"
	find_callback_record $record_list "bpm_y_kd" "bpm_y_kd"

	start_feedback_control_callbacks
}

############################################################################

proc create_bpm_frame { record_list frame } {

	package require BLT

	set bpm_plot $frame.bpm_plot

	set FeedbackControl::bpm_plot $bpm_plot

	blt::graph $bpm_plot -plotbackground white

	$bpm_plot element create beamspot

	# Configure the BPM plot

	$bpm_plot configure -title "Beam Position Monitor"

	$bpm_plot element configure beamspot -label ""
	$bpm_plot element configure beamspot -color red

	$bpm_plot axis configure x -min -4000.0 -max 4000.0
	$bpm_plot axis configure y -min -4000.0 -max 4000.0

	$bpm_plot axis configure x -title "x (um)"
	$bpm_plot axis configure y -title "y (um)"

	$bpm_plot grid configure -hide 0

	pack $bpm_plot -side left

	# Display the x and y positions

	set xyframe $frame.xyframe

	frame $xyframe
	pack $xyframe -side left

	set row 0

	label $xyframe.x_label -text "X:"
	label $xyframe.x_value -text "----" -fg blue
	label $xyframe.x_units -text "um"

	grid $xyframe.x_label -row $row -column 0
	grid $xyframe.x_value -row $row -column 1
	grid $xyframe.x_units -row $row -column 2

	incr row

	label $xyframe.y_label -text "Y:"
	label $xyframe.y_value -text "----" -fg blue
	label $xyframe.y_units -text "um"

	grid $xyframe.y_label -row $row -column 0
	grid $xyframe.y_value -row $row -column 1
	grid $xyframe.y_units -row $row -column 2

	incr row

	label $xyframe.spacer1 -text "   "
	grid $xyframe.spacer1 -row $row -column 1 -pady 5

	incr row

	button $xyframe.x_enable -bg gray -text "X Feedback" \
	    -command [ itcl::code change_feedback $record_list "x_fb_enable" ]
	grid $xyframe.x_enable -row $row -column 1

	incr row

	label $xyframe.spacer1a -text "   "
	grid $xyframe.spacer1a -row $row -column 1 -pady 2

	incr row

	button $xyframe.y_enable -bg gray -text "Y Feedback" \
	    -command [ itcl::code change_feedback $record_list "y_fb_enable" ]
	grid $xyframe.y_enable -row $row -column 1

	incr row

	label $xyframe.spacer2 -text "   "
	grid $xyframe.spacer2 -row $row -column 1 -pady 5

	incr row

	label $xyframe.chi_label -text "Chi:"
	label $xyframe.chi_value -fg blue -text "----"
	label $xyframe.chi_units -text "urad"

	grid $xyframe.chi_label -row $row -column 0 -sticky e
	grid $xyframe.chi_value -row $row -column 1
	grid $xyframe.chi_units -row $row -column 2

	incr row

	label $xyframe.mjack_pitch_label -text "Mjack_pitch:"
	label $xyframe.mjack_pitch_value -fg blue -text "----"
	label $xyframe.mjack_pitch_units -text "urad"

	grid $xyframe.mjack_pitch_label -row $row -column 0 -sticky e
	grid $xyframe.mjack_pitch_value -row $row -column 1
	grid $xyframe.mjack_pitch_units -row $row -column 2

	incr row

	label $xyframe.spacer3 -text "   "
	grid $xyframe.spacer3 -row $row -column 1 -pady 5

	incr row

	label $xyframe.bpm_sum_label -text "BPM Sum:"
	label $xyframe.bpm_sum_value -fg blue -text "----"

	grid $xyframe.bpm_sum_label -row $row -column 0 -sticky e
	grid $xyframe.bpm_sum_value -row $row -column 1

	incr row

	label $xyframe.spacer4 -text "   "
	grid $xyframe.spacer4 -row $row -column 1 -pady 5

	incr row

	label $xyframe.bpm_curr1_label -text "Current 1:"
	label $xyframe.bpm_curr1_value -fg blue -text "----"

	grid $xyframe.bpm_curr1_label -row $row -column 0 -sticky e
	grid $xyframe.bpm_curr1_value -row $row -column 1

	label $xyframe.bpm_curr2_label -text "Current 2:"
	label $xyframe.bpm_curr2_value -fg blue -text "----"

	grid $xyframe.bpm_curr2_label -row $row -column 2 -sticky e
	grid $xyframe.bpm_curr2_value -row $row -column 3

	incr row

	label $xyframe.bpm_curr3_label -text "Current 3:"
	label $xyframe.bpm_curr3_value -fg blue -text "----"

	grid $xyframe.bpm_curr3_label -row $row -column 0 -sticky e
	grid $xyframe.bpm_curr3_value -row $row -column 1

	label $xyframe.bpm_curr4_label -text "Current 4:"
	label $xyframe.bpm_curr4_value -fg blue -text "----"

	grid $xyframe.bpm_curr4_label -row $row -column 2 -sticky e
	grid $xyframe.bpm_curr4_value -row $row -column 3
}

proc change_feedback { record_list record_name } {
	global callback_values

	set feedback_enable $callback_values($record_name)

	if { $feedback_enable } {
		set feedback_enable 0
	} else {
		set feedback_enable 1
	}

	set feedback_record [ $record_list get_record $record_name ]

	$feedback_record send_variable $feedback_enable
}

############################################################################

proc update_bpm_frame { frame bpm } {
	global callback_values

#	log_puts "update_bpm_frame invoked, frame = $frame, bpm = $bpm"

	set bpm_posx $callback_values(bpm_posx)

#	set x0 [ lindex $callback_values(bpm_x_params) 0 ]
#	set x1 [ lindex $callback_values(bpm_x_params) 1 ]
#	set x2 [ lindex $callback_values(bpm_x_params) 2 ]
#	set x3 [ lindex $callback_values(bpm_x_params) 3 ]
#
#	set bpm_x [ expr ( $x0 + $x1 * $bpm_posx \
#			+ $x2 * $bpm_posx * $bpm_posx \
#			+ $x3 * $bpm_posx * $bpm_posx * $bpm_posx ) ]

	set bpm_x $bpm_posx

	set bpm_posy $callback_values(bpm_posy)

#	set y0 [ lindex $callback_values(bpm_y_params) 0 ]
#	set y1 [ lindex $callback_values(bpm_y_params) 1 ]
#	set y2 [ lindex $callback_values(bpm_y_params) 2 ]
#	set y3 [ lindex $callback_values(bpm_y_params) 3 ]
#
#	set bpm_y [ expr ( $y0 + $y1 * $bpm_posy \
#			+ $y2 * $bpm_posy * $bpm_posy \
#			+ $y3 * $bpm_posy * $bpm_posy * $bpm_posy ) ]

	set bpm_y $bpm_posy

	$frame.bpm_plot element configure beamspot -xdata $bpm_x

	$frame.bpm_plot element configure beamspot -ydata $bpm_y

	$frame.xyframe.x_value configure -text $bpm_x
	$frame.xyframe.y_value configure -text $bpm_y

	set x_fb_enable $callback_values(x_fb_enable)

	set button $frame.xyframe.x_enable

	if { $x_fb_enable } {
		$button configure -text "X Feedback\nEnabled"
		$button configure -bg limegreen -activebackground limegreen
	} else {
		$button configure -text "X Feedback\nDisabled"
		$button configure -bg red -activebackground red
	}

	set y_fb_enable $callback_values(y_fb_enable)

	set button $frame.xyframe.y_enable

	if { $y_fb_enable } {
		$button configure -text "Y Feedback\nEnabled"
		$button configure -bg limegreen -activebackground limegreen
	} else {
		$button configure -text "Y Feedback\nDisabled"
		$button configure -bg red -activebackground red
	}

	$frame.xyframe.mjack_pitch_value \
			configure -text $callback_values(mjack_pitch)
	$frame.xyframe.chi_value   configure -text $callback_values(chi)

	$frame.xyframe.bpm_sum_value \
			configure -text $callback_values(bpm_sum)

	$frame.xyframe.bpm_curr1_value \
			configure -text $callback_values(bpm_curr1)

	$frame.xyframe.bpm_curr2_value \
			configure -text $callback_values(bpm_curr2)

	$frame.xyframe.bpm_curr3_value \
			configure -text $callback_values(bpm_curr3)

	$frame.xyframe.bpm_curr4_value \
			configure -text $callback_values(bpm_curr4)
}

############################################################################

proc create_setpoint_frame { record_list frame } {

	label $frame.title -text "Setpoint Control"
	pack $frame.title -side top

	iwidgets::labeledframe $frame.sp_frame \
			-labelpos nw -labeltext "Setpoints"
	pack $frame.sp_frame -side top

	label $frame.sp_frame.x_label -text "X Setpoint:"
	label $frame.sp_frame.x_value -fg blue -text "----"

	iwidgets::entryfield $frame.sp_frame.x_setpoint -width 15 \
		-command [ itcl::code set_setpoint_value $record_list \
			"bpm_x_setpoint" $frame.sp_frame.x_setpoint ]

###		-validate { iwidgets::Entryfield::real %P } 

	grid $frame.sp_frame.x_label    -row 0 -column 0 -padx 3 -pady 5
	grid $frame.sp_frame.x_value    -row 0 -column 1 -padx 3 -pady 5
	grid $frame.sp_frame.x_setpoint -row 0 -column 2 -padx 3 -pady 5

	label $frame.sp_frame.spacer -text "    "
	grid $frame.sp_frame.spacer     -row 0 -column 3 -padx 3 -pady 5

	label $frame.sp_frame.y_label -text "Y Setpoint:"
	label $frame.sp_frame.y_value -fg blue -text "----"

	iwidgets::entryfield $frame.sp_frame.y_setpoint -width 15 \
		-command [ itcl::code set_setpoint_value $record_list \
			"bpm_y_setpoint" $frame.sp_frame.y_setpoint ]

	grid $frame.sp_frame.y_label    -row 0 -column 4 -padx 3 -pady 5
	grid $frame.sp_frame.y_value    -row 0 -column 5 -padx 3 -pady 5
	grid $frame.sp_frame.y_setpoint -row 0 -column 6 -padx 3 -pady 5
}

############################################################################

proc update_setpoint_frame { frame } {
	global callback_values

#	log_puts "update_setpoint_frame invoked, frame = $frame"

	$frame.sp_frame.x_value configure -text $callback_values(bpm_x_setpoint)
	$frame.sp_frame.y_value configure -text $callback_values(bpm_y_setpoint)
}

############################################################################

proc set_setpoint_value { record_list record_name entryfield } {

	set record [ $record_list get_record $record_name ]

	set value [ $entryfield get ]

	log_puts "record = $record, value = $value"

	$record send_variable $value
}

############################################################################

proc create_servo_frame { record_list frame } {

	label $frame.title -text "Servo Control"
	pack $frame.title -side top

	iwidgets::labeledframe $frame.x \
		-labelpos nw -labeltext "BPM X"
	pack $frame.x -side top

	fill_servo_labeled_frame $record_list $frame.x "x"

	iwidgets::labeledframe $frame.y \
		-labelpos nw -labeltext "BPM Y"
	pack $frame.y -side top

	fill_servo_labeled_frame $record_list $frame.y "y"
}

proc fill_servo_labeled_frame { record_list grid axis } {

	fill_servo_gain_fields $record_list $grid 0 $axis "kp" "Kp:"
	fill_servo_gain_fields $record_list $grid 3 $axis "ki" "Ki:"
	fill_servo_gain_fields $record_list $grid 6 $axis "kd" "Kd:"
}

proc fill_servo_gain_fields { record_list grid col axis gain gain_label } {

	label $grid.${gain}_label -text $gain_label
	label $grid.${gain}_value -fg blue -text "----"

	iwidgets::entryfield $grid.${gain}_setpoint -width 6 \
		-validate { iwidgets::Entryfield::real %P } \
		-command [ itcl::code set_servo_value $record_list \
			"bpm_${axis}_${gain}" $grid.${gain}_setpoint ]

	grid $grid.${gain}_label    -row 0 -column $col -padx 3 -pady 5
	incr col
	grid $grid.${gain}_value    -row 0 -column $col -padx 3 -pady 5
	incr col
	grid $grid.${gain}_setpoint -row 0 -column $col -padx 3 -pady 5
}

############################################################################

proc update_servo_frame { frame } {
	global callback_values

#	log_puts "update_servo_frame invoked, frame = $frame"

	$frame.x.kp_value configure -text $callback_values(bpm_x_kp)
	$frame.x.ki_value configure -text $callback_values(bpm_x_ki)
	$frame.x.kd_value configure -text $callback_values(bpm_x_kd)

	$frame.y.kp_value configure -text $callback_values(bpm_y_kp)
	$frame.y.ki_value configure -text $callback_values(bpm_y_ki)
	$frame.y.kd_value configure -text $callback_values(bpm_y_kd)
}

############################################################################

proc set_servo_value { record_list record_name entryfield } {

	set record [ $record_list get_record $record_name ]

	set value [ $entryfield get ]

#	log_puts "record = $record, value = $value"

	$record send_variable $value
}

############################################################################

proc create_param_frame { record_list frame } {

	label $frame.title -text "Fit Parameters"
	pack $frame.title -side top

	set grid_xp $frame.grid_xp

	iwidgets::labeledframe $grid_xp \
		-labelpos nw -labeltext "X Fit Parameters"

	pack $grid_xp -side left

	label $grid_xp.value -fg blue -text "----"
	grid $grid_xp.value -row 0 -column 0 -padx 3 -pady 5

	iwidgets::entryfield $grid_xp.setpoint -width 15 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_x_params" $grid_xp.setpoint ]

	grid $grid_xp.setpoint -row 0 -column 1 -padx 3 -pady 5

	#-------------------

	set grid_yp $frame.grid_yp

	iwidgets::labeledframe $grid_yp \
		-labelpos nw -labeltext "Y Fit Parameters"

	pack $grid_yp -side left

	label $grid_yp.value -fg blue -text "----"
	grid $grid_yp.value -row 0 -column 0 -padx 3 -pady 5

	iwidgets::entryfield $grid_yp.setpoint -width 15 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_y_params" $grid_yp.setpoint ]

	grid $grid_yp.setpoint -row 0 -column 1 -padx 3 -pady 5
}

proc update_param_frame { frame } {
	global callback_values

	$frame.grid_xp.value configure -text $callback_values(bpm_x_params)
	$frame.grid_yp.value configure -text $callback_values(bpm_y_params)
}

############################################################################

proc create_limits_frame { record_list frame } {

	label $frame.title -text "Feedback Limits"
	pack $frame.title -side top

	#===================

	set grid_sum $frame.grid_sum

	iwidgets::labeledframe $grid_sum \
		-labelpos nw -labeltext "Readout Limits"

	pack $grid_sum -side left

	#-------------------

	label $grid_sum.sum_label -text "BPM Sum Min:"
	grid $grid_sum.sum_label -row 0 -column 0 -padx 3 -pady 5

	label $grid_sum.sum_value -fg blue -text "----"
	grid $grid_sum.sum_value -row 0 -column 1 -padx 3 -pady 5

	iwidgets::entryfield $grid_sum.sum_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_sum_min" $grid_sum.sum_setpoint ]

	grid $grid_sum.sum_setpoint -row 0 -column 2 -padx 3 -pady 5

	#===================

	set grid_lim $frame.grid_lim

	iwidgets::labeledframe $grid_lim \
		-labelpos nw -labeltext "Position Limits"

	pack $grid_lim -side left

	#-------------------

	label $grid_lim.x_label -text "Chi:"
	grid $grid_lim.x_label -row 0 -column 0 -padx 3 -pady 5 -sticky e

	#-------------------

	label $grid_lim.xmin_label -text "Min:"
	grid $grid_lim.xmin_label -row 0 -column 1 -padx 3 -pady 5

	label $grid_lim.xmin_value -fg blue -text "----"
	grid $grid_lim.xmin_value -row 0 -column 2 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.xmin_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_x_min" $grid_lim.xmin_setpoint ]

	grid $grid_lim.xmin_setpoint -row 0 -column 3 -padx 3 -pady 5

	#-------------------

	label $grid_lim.xmax_label -text "Max:"
	grid $grid_lim.xmax_label -row 0 -column 4 -padx 3 -pady 5

	label $grid_lim.xmax_value -fg blue -text "----"
	grid $grid_lim.xmax_value -row 0 -column 5 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.xmax_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_x_max" $grid_lim.xmax_setpoint ]

	grid $grid_lim.xmax_setpoint -row 0 -column 6 -padx 3 -pady 5

	#-------------------

	label $grid_lim.xdb_label -text "Move Deadband:"
	grid $grid_lim.xdb_label -row 0 -column 7 -padx 3 -pady 5

	label $grid_lim.xdb_value -fg blue -text "----"
	grid $grid_lim.xdb_value -row 0 -column 8 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.xdb_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_x_deadband" $grid_lim.xdb_setpoint ]

	grid $grid_lim.xdb_setpoint -row 0 -column 9 -padx 3 -pady 5

	#-------------------

	label $grid_lim.y_label -text "Mjack_pitch:"
	grid $grid_lim.y_label -row 1 -column 0 -padx 3 -pady 5 -sticky e

	#-------------------

	label $grid_lim.ymin_label -text "Min:"
	grid $grid_lim.ymin_label -row 1 -column 1 -padx 3 -pady 5

	label $grid_lim.ymin_value -fg blue -text "----"
	grid $grid_lim.ymin_value -row 1 -column 2 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.ymin_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_y_min" $grid_lim.ymin_setpoint ]

	grid $grid_lim.ymin_setpoint -row 1 -column 3 -padx 3 -pady 5

	#-------------------

	label $grid_lim.ymax_label -text "Max:"
	grid $grid_lim.ymax_label -row 1 -column 4 -padx 3 -pady 5

	label $grid_lim.ymax_value -fg blue -text "----"
	grid $grid_lim.ymax_value -row 1 -column 5 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.ymax_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_y_max" $grid_lim.ymax_setpoint ]

	grid $grid_lim.ymax_setpoint -row 1 -column 6 -padx 3 -pady 5

	#-------------------

	label $grid_lim.ydb_label -text "Move Deadband:"
	grid $grid_lim.ydb_label -row 1 -column 7 -padx 3 -pady 5

	label $grid_lim.ydb_value -fg blue -text "----"
	grid $grid_lim.ydb_value -row 1 -column 8 -padx 3 -pady 5

	iwidgets::entryfield $grid_lim.ydb_setpoint -width 6 \
		-command [ itcl::code set_qbpm_value $record_list \
			"bpm_y_deadband" $grid_lim.ydb_setpoint ]

	grid $grid_lim.ydb_setpoint -row 1 -column 9 -padx 3 -pady 5

	#-------------------
}

proc update_limits_frame { frame } {
	global callback_values

	$frame.grid_sum.sum_value configure -text $callback_values(bpm_sum_min)

	$frame.grid_lim.xmin_value configure -text $callback_values(bpm_x_min)
	$frame.grid_lim.xmax_value configure -text $callback_values(bpm_x_max)
	$frame.grid_lim.xdb_value configure -text \
					$callback_values(bpm_x_deadband)

	$frame.grid_lim.ymin_value configure -text $callback_values(bpm_y_min)
	$frame.grid_lim.ymax_value configure -text $callback_values(bpm_y_max)
	$frame.grid_lim.ydb_value configure -text \
					$callback_values(bpm_y_deadband)
}

############################################################################

proc create_qbpm_control_frame { record_list bpm frame } {

	label $frame.title -text "QBPM Control"
	pack $frame.title -side top

	set grid $frame.grid

	iwidgets::labeledframe $grid \
		-labelpos nw -labeltext "Downstream BPM"
	pack $grid -side top

	fill_qbpm_labeled_frame $record_list $grid "${bpm}"

	set grid_average $frame.grid_average

	iwidgets::labeledframe $grid_average \
		-labelpos nw -labeltext "Averaging"
	pack $grid_average -side top

	fill_qbpm_average_frame $record_list $bpm $grid_average
}

proc fill_qbpm_labeled_frame { record_list grid bpm } {

	create_qbpm_column $record_list $grid $bpm 0 \
					"gain" "offset" "Gain:" "Offset:"
	create_qbpm_column $record_list $grid $bpm 3 "gx" "gy" "GX:" "GY:"
	create_qbpm_column $record_list $grid $bpm 6 "a1" "a2" "A1:" "A2:"
	create_qbpm_column $record_list $grid $bpm 9 "b1" "b2" "B1:" "B2:"
	create_qbpm_column $record_list $grid $bpm 12 "c1" "c2" "C1:" "C2:"
	create_qbpm_column $record_list $grid $bpm 15 "d1" "d2" "D1:" "D2:"
}

proc create_qbpm_column { record_list grid bpm \
				col top bottom top_text bottom_text } {
	label $grid.${top}_label -text $top_text
	label $grid.${bottom}_label -text $bottom_text

	grid $grid.${top}_label -row 0 -column $col -padx 3 -pady 5
	grid $grid.${bottom}_label -row 1 -column $col -padx 3 -pady 5

	incr col

	label $grid.${top}_value -fg blue -text "----" 
	label $grid.${bottom}_value -fg blue -text "----" 

	grid $grid.${top}_value -row 0 -column $col -padx 3 -pady 5
	grid $grid.${bottom}_value -row 1 -column $col -padx 3 -pady 5

	incr col

	iwidgets::entryfield $grid.${top}_setpoint -width 6 \
		-validate { iwidgets::Entryfield::real %P } \
		-command [ itcl::code set_qbpm_value $record_list \
			"${bpm}_${top}" $grid.${top}_setpoint ]

	iwidgets::entryfield $grid.${bottom}_setpoint -width 6 \
		-validate { iwidgets::Entryfield::real %P } \
		-command [ itcl::code set_qbpm_value $record_list \
			"${bpm}_${bottom}" $grid.${bottom}_setpoint ]

	grid $grid.${top}_setpoint -row 0 -column $col -padx 3 -pady 5
	grid $grid.${bottom}_setpoint -row 1 -column $col -padx 3 -pady 5
}

proc fill_qbpm_average_frame { record_list bpm grid } {

	label $grid.ds_label -text "In QBPM:"
	label $grid.ds_value -fg blue -text "----"

	iwidgets::entryfield $grid.ds_setpoint -width 6 \
		-validate { iwidgets::Entryfield::integer %P } \
		-command [ itcl::code set_qbpm_value $record_list \
			"${bpm}_average" $grid.ds_setpoint ]

	grid $grid.ds_label    -row 0 -column 0 -padx 3 -pady 5
	grid $grid.ds_value    -row 0 -column 1 -padx 3 -pady 5
	grid $grid.ds_setpoint -row 0 -column 2 -padx 3 -pady 5

	label $grid.spacer -text "    "
	grid $grid.spacer -row 0 -column 3 -padx 3 -pady 5

	label $grid.fbavg_label -text "In Feedback Script:"
	label $grid.fbavg_value -fg blue -text "---"

	iwidgets::entryfield $grid.fbavg_setpoint -width 6 \
		-validate { iwidgets::Entryfield::integer %P } \
		-command [ itcl::code set_qbpm_value $record_list \
			"num_qbpm_reads" $grid.fbavg_setpoint ]

	grid $grid.fbavg_label    -row 0 -column 4 -padx 3 -pady 5
	grid $grid.fbavg_value    -row 0 -column 5 -padx 3 -pady 5
	grid $grid.fbavg_setpoint -row 0 -column 6 -padx 3 -pady 5
}

############################################################################

proc update_qbpm_control_frame { frame } {
	global callback_values

#	log_puts "update_qbpm_control_frame invoked, frame = $frame"

	set grid $frame.grid

	$grid.gain_value configure -text \
		[ feedback_format "%.3e" $callback_values(bpm_gain) ]
	$grid.offset_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_offset) ]
	$grid.gx_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_gx) ]
	$grid.gy_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_gy) ]
	$grid.a1_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_a1) ]
	$grid.a2_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_a2) ]
	$grid.b1_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_b1) ]
	$grid.b2_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_b2) ]
	$grid.c1_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_c1) ]
	$grid.c2_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_c2) ]
	$grid.d1_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_d1) ]
	$grid.d2_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_d2) ]

	set grid_average $frame.grid_average

	$grid_average.ds_value configure -text \
		[ feedback_format "%g" $callback_values(bpm_average) ]
	$grid_average.fbavg_value configure -text \
		[ feedback_format "%g" $callback_values(num_qbpm_reads) ]
}

############################################################################

proc set_qbpm_value { record_list record_name entryfield } {

	set value [ $entryfield get ]

	if { $record_name == "bpm_us_gain" } {
		set record [ $record_list get_record "bpm_us" ]

		$record set_gain $value

	} elseif { $record_name == "bpm_us_offset" } {
		set record [ $record_list get_record "bpm_us" ]

		$record set_offset $value

	} elseif { $record_name == "bpm_ds_gain" } {
		set record [ $record_list get_record "bpm_ds" ]

		$record set_gain $value

	} elseif { $record_name == "bpm_ds_offset" } {
		set record [ $record_list get_record "bpm_ds" ]

		$record set_offset $value

	} elseif { $record_name == "bpm_x_params" } {
		set record [ $record_list get_record "bpm_x_params" ]

		$record send_variable $value

	} elseif { $record_name == "bpm_y_params" } {
		set record [ $record_list get_record "bpm_y_params" ]

		$record send_variable $value

	} else {
		set record [ $record_list get_record $record_name ]

		$record write $value
	}

#	log_puts "record = $record, value = $value"
}

